/**
 * Copyright (c) 2014-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.junit.rules.webserver;

import static org.junit.Assert.assertNotNull;

import java.net.URL;

import javax.servlet.http.HttpServlet;

import org.eclipse.jetty.http.HttpVersion;
import org.eclipse.jetty.server.HttpConfiguration;
import org.eclipse.jetty.server.HttpConnectionFactory;
import org.eclipse.jetty.server.SecureRequestCustomizer;
import org.eclipse.jetty.server.Server;
import org.eclipse.jetty.server.ServerConnector;
import org.eclipse.jetty.server.SslConnectionFactory;
import org.eclipse.jetty.servlet.ServletHandler;
import org.eclipse.jetty.servlet.ServletHolder;
import org.eclipse.jetty.util.ssl.SslContextFactory;
import org.junit.Rule;
import org.junit.rules.ExternalResource;

/**
 * The {@link WebServer} {@link Rule} allows creation of a web server resource. It is guaranteed to be deleted when the
 * test method finishes (whether it passes or fails):
 * 
 * <pre>
 * public class MyTestSuite {
 *  &#064;Rule
 *  public WebServer webServer = new WebServer(...);
 * 
 *  &#064;Test
 *  public void test...() throws IOException {
 *      webServer.addServlet(servlet, path);
 *      ...
 *  }
 * }
 * </pre>
 */
// TODO: Reuse this rule at Petals container level for tests of DownloadTask
public class WebServer extends ExternalResource {

    /**
     * The default HTTP port
     */
    public static final int DEFAULT_HTTP_PORT = 10080;

    /**
     * The port on which HTTP requests are listen
     */
    private final int httpPort;

    /**
     * The web server
     */
    private final Server server;

    /**
     * Servlet handler of the web server
     */
    private final ServletHandler servletHandler;

    /**
     * Creates a temporary web-server listening HTTP requests and HTTPS requests on the given ports.
     * 
     * @param httpPort
     *            The HTTP port on which requests will be listen. Not {@code null} if {@code httpsPort} is {@code null}.
     * @param httpsPort
     *            The HTTPS port on which requests will be listen. Not {@code null} if {@code httpPort} is {@code null}.
     * @param keystoreResourceName
     *            Keystore file as resource name. Not {@code null} if {@code httpsPort} is not {@code null}.
     * @param keystorePassword
     *            The keystore password. Not {@code null} if {@code httpsPort} is not {@code null}.
     * @param keyManagerPassword
     *            The password (if any) for the specific key within the key store. Can be {@code null}.
     */
    public WebServer(final int httpPort, final int httpsPort, final String keystoreResourceName,
            final String keystorePassword, final String keyManagerPassword) {

        assert httpPort != 0 || httpsPort != 0;

        this.httpPort = httpPort;

        this.server = new Server();

        if (httpsPort != 0) {

            assert keystoreResourceName != null;
            assert keystorePassword != null;

            final HttpConfiguration httpsCfg = new HttpConfiguration();
            httpsCfg.addCustomizer(new SecureRequestCustomizer());

            final SslContextFactory sslContextFactory = new SslContextFactory();
            final URL keystoreUrl = Thread.currentThread().getContextClassLoader().getResource(keystoreResourceName);
            assertNotNull(keystoreUrl);
            sslContextFactory.setKeyStorePath(
                    keystoreUrl.toExternalForm());
            sslContextFactory.setKeyStorePassword(keystorePassword);
            sslContextFactory.setKeyManagerPassword(keyManagerPassword);

            final ServerConnector sslConnector = new ServerConnector(this.server,
                    new SslConnectionFactory(sslContextFactory, HttpVersion.HTTP_1_1.asString()),
                    new HttpConnectionFactory(httpsCfg));
            sslConnector.setPort(httpsPort);

            this.server.addConnector(sslConnector);
        }

        if (httpPort != 0) {
            final ServerConnector httpConnector = new ServerConnector(this.server);
            httpConnector.setPort(httpPort);
            this.server.addConnector(httpConnector);
        }

        this.servletHandler = new ServletHandler();
        this.server.setHandler(this.servletHandler);

    }

    /**
     * Creates a temporary web-server listening HTTP requests on the given port.
     * 
     * @param httpPort
     *            The HTTP port on which requests will be listen
     */
    public WebServer(final int httpPort) {
        this(httpPort, 0, null, null, null);
    }

    /**
     * Creates a temporary web-server listening HTTP requests on the default port ({@value #DEFAULT_HTTP_PORT})
     */
    public WebServer() {
        this(DEFAULT_HTTP_PORT);
    }

    @Override
    protected void before() throws Throwable {
        this.initializeWebServer();
    }

    @Override
    protected void after() {
        try {
            this.destroyWebServer();
        } catch (final Exception e) {
            e.printStackTrace();
        }
    }

    /**
     * Initialize the Web server
     */
    private void initializeWebServer() throws Exception {

        this.server.start();
    }

    /**
     * Free the web server
     */
    private void destroyWebServer() throws Exception {

        this.server.stop();
    }

    /**
     * @return The beginning of URL served by the web server: <code>http://&lt;server&gt;:&lt;port&gt;</code>
     */
    public String getHttpBaseUrl() {
        return "http://localhost:" + this.httpPort;
    }

    /**
     * Add a servlet to the web server
     * 
     * @param httpServlet
     *            The servlet to add
     * @param path
     *            The path of the URL to access the servlet
     */
    public void addServlet(final HttpServlet httpServlet, final String path) {

        this.servletHandler.addServletWithMapping(new ServletHolder(httpServlet), path);
    }

    /**
     * Registering a servlet into the web-server in a chaining way.
     * 
     * @param httpServlet
     *            The servlet to add
     * @param path
     *            The path of the URL to access the servlet
     * @return
     */
    public WebServer withServlet(final HttpServlet httpServlet, final String path) {

        this.addServlet(httpServlet, path);
        return this;
    }

}
